# encoding: UTF-8

Encoding.default_internal = Encoding.default_external = 'UTF-8'

require 'bigdecimal'
require 'logger'

require 'time'

require 'rubygems'

require 'multi_json'

MultiJson.engine = :json_gem

require 'subledger/exception_handler'

module Subledger
  class Error < StandardError; end
end

require 'subledger/collection_name'
require 'subledger/domain'
require 'subledger/interface'
require 'subledger/path'
require 'subledger/rest'
require 'subledger/uuid'
require 'subledger/version'

# additional requires at bottom

SUBLEDGER = 'SUBLEDGER'

module Subledger
  class DateError < Error; end

  ROOT = File.expand_path __FILE__ + '../../..'

  ROOT_CONFIG = ROOT + '/config'
  ROOT_LIB    = ROOT + '/lib'
  ROOT_SPEC   = ROOT + '/spec'

  LOG = ::Logger.new '/tmp/subledger.log', 10, 102_400_000
  LOG.sev_threshold = Logger::DEBUG

  case ::ENV['SUBLEDGER_ENV']
    when nil
      ENV    = 'production'
      COVER  = ::ENV['SUBLEDGER_COVER' ] || false
      FORMAT = ::ENV['SUBLEDGER_FORMAT'] || 'pride'
      SCHEME = ::ENV['SUBLEDGER_SCHEME'] || 'https://'
      DOMAIN = ::ENV['SUBLEDGER_DOMAIN'] || 'api.subledger.com'

      STORE  = ( ::ENV['SUBLEDGER_STORE'] || :api ).to_sym
      AWS_CRED_FILE = nil

      LOG.sev_threshold = Logger::WARN

    when 'test'
      ENV    = 'test'
      COVER  = ::ENV['SUBLEDGER_COVER' ] || false
      FORMAT = ::ENV['SUBLEDGER_FORMAT'] || 'pride'
      SCHEME = ::ENV['SUBLEDGER_SCHEME'] || 'http://'
      DOMAIN = ::ENV['SUBLEDGER_DOMAIN'] || 'api.subledger.dev'

      STORE  = ( ::ENV['SUBLEDGER_STORE'] || :memory ).to_sym
      AWS_CRED_FILE = 'aws_creds_dev.yml'

    when 'test_single'
      ENV    = 'test'
      COVER  = ::ENV['SUBLEDGER_COVER' ] || false
      FORMAT = ::ENV['SUBLEDGER_FORMAT'] || 'turn'
      SCHEME = ::ENV['SUBLEDGER_SCHEME'] || 'http://'
      DOMAIN = ::ENV['SUBLEDGER_DOMAIN'] || 'api.subledger.dev'

      STORE  = ( ::ENV['SUBLEDGER_STORE'] || :memory ).to_sym
      AWS_CRED_FILE = 'aws_creds_dev.yml'

    when 'ci'
      ENV    = 'ci'
      COVER  = ::ENV['SUBLEDGER_COVER' ] || false
      FORMAT = ::ENV['SUBLEDGER_FORMAT'] || 'normal'
      SCHEME = ::ENV['SUBLEDGER_SCHEME'] || 'http://'
      DOMAIN = ::ENV['SUBLEDGER_DOMAIN'] || 'localhost:9292'

      STORE  = ( ::ENV['SUBLEDGER_STORE'] || :memory ).to_sym
      AWS_CRED_FILE = 'aws_creds_dev.yml'

    when 'production'
      ENV    = 'production'
      COVER  = ::ENV['SUBLEDGER_COVER' ] || false
      FORMAT = ::ENV['SUBLEDGER_FORMAT'] || 'pride'
      SCHEME = ::ENV['SUBLEDGER_SCHEME'] || 'https://'
      DOMAIN = ::ENV['SUBLEDGER_DOMAIN'] || 'api.subledger.com'

      STORE  = ( ::ENV['SUBLEDGER_STORE'] || :aws ).to_sym
      AWS_CRED_FILE = 'aws_creds_prod.yml'
  end

  case STORE
    when :memory
      SLOW_POST_DELAY      = ::ENV['SUBLEDGER_SLOW_POST_DELAY']      || 1
      PROGRESS_INITIAL     = ::ENV['SUBLEDGER_PROGRESS_INITIAL']     || 0.075
      PROGRESS_RETRY       = ::ENV['SUBLEDGER_PROGRESS_RETRY']       || 0.05
      PROGRESS_FAIL        = ::ENV['SUBLEDGER_PROGRESS_FAIL']        || 4
      BALANCE_SETTLE_DELAY = ::ENV['SUBLEDGER_BALANCE_SETTLE_DELAY'] || 0
    when :api
      SLOW_POST_DELAY      = ::ENV['SUBLEDGER_SLOW_POST_DELAY']      || 1
      PROGRESS_INITIAL     = ::ENV['SUBLEDGER_PROGRESS_INITIAL']     || 0.075
      PROGRESS_RETRY       = ::ENV['SUBLEDGER_PROGRESS_RETRY']       || 0.05
      PROGRESS_FAIL        = ::ENV['SUBLEDGER_PROGRESS_FAIL']        || 4
      BALANCE_SETTLE_DELAY = ::ENV['SUBLEDGER_BALANCE_SETTLE_DELAY'] || 0
    when :aws
      SLOW_POST_DELAY      = ::ENV['SUBLEDGER_SLOW_POST_DELAY']      || 15
      PROGRESS_INITIAL     = ::ENV['SUBLEDGER_PROGRESS_INITIAL']     || 1.5
      PROGRESS_RETRY       = ::ENV['SUBLEDGER_PROGRESS_RETRY']       || 0.5
      PROGRESS_FAIL        = ::ENV['SUBLEDGER_PROGRESS_FAIL']        || 20
      BALANCE_SETTLE_DELAY = ::ENV['SUBLEDGER_BALANCE_SETTLE_DELAY'] || 3
  end

  unless AWS_CRED_FILE.nil?
    AWS_PATH = File.expand_path( File.dirname( __FILE__ ) + '/../config/' + AWS_CRED_FILE )
  end

  RENDERED_REPORTS_BUCKET           = "subledger_#{ENV}_rendered_reports"
  RENDERED_REPORT_CATEGORIES_BUCKET = "subledger_#{ENV}_rendered_report_categories"
  BACKUP_BUCKET                     = "subledger_#{ENV}_backups"

  def self.new args = { }
    args = args.dup

    # auth args

    identity    = args[:identity]
    identity_id = args[:identity_id]

    if identity.nil? and not identity_id.nil?
      args.merge! :identity => Domain::Identity.
                                 new( :id => identity_id )

      args.delete :identity_id
    end

    key    = args[:key]
    key_id = args[:key_id]

    if key.nil? and not key_id.nil?
      args.merge! :auth_key => Domain::Key.
                                 new( :id       => key_id,
                                      :identity => args[:identity],
                                      :secret   => args[:secret] )

      args.delete :key_id
      args.delete :secret
    end

    # instance args

    org    = args[:org]
    org_id = args[:org_id]

    if org.nil? and not org_id.nil?
      args.merge! :org => Domain::Org.
                            new( :id => org_id )

      args.delete :org_id
    end

    book    = args[:book]
    book_id = args[:book_id]

    if book.nil? and not book_id.nil?
      args.merge! :book => Domain::Book.
                             new( :id  => book_id,
                                  :org => args[:org] )

      args.delete :book_id
    end

    args.merge! :store => store( args )

    Interface::Client.new args
  end

  def self.store args = { }
    store_sym = args[:store] || STORE

    begin
      require "subledger/store/#{store_sym}"
    rescue LoadError => e
      raise Error, "Store: #{store_sym}: #{e}"
    end

    Object.const_get(       'Subledger'          ).
             const_get(     'Store'                ).
               const_get(    store_sym.capitalize    ).
                 const_get( 'Store'                    ).new args
  end

  # Thanks to Jan Goyvaerts, Steven Levithan, O'Reilly Media, Inc.
  ISO8601_REGEXP = /^(?<year>[0-9]{4})-?(?<month>1[0-2]|0[1-9])-?(?<day>3[0-1]|0[1-9]|[1-2][0-9])T(?<hour>2[0-3]|[0-1][0-9]):?(?<minute>[0-5][0-9]):?(?<second>[0-5][0-9])(?<ms>\.[0-9]{1,3})??(?<timezone>Z|[+-](?:2[0-3]|[0-1][0-9]):?[0-5][0-9])$/
  # http://my.safaribooksonline.com/book/programming/regular-expressions/9780596802837/4dot-validation-and-formatting/id2983571
  # Modified to accept optional punctuation and fractional seconds of 1 to 3 digits

  EARLIEST_TIME = Time.utc 1970, 1, 1, 0, 0, 0, 0
  LATEST_TIME   = EARLIEST_TIME + 2 ** 32

  def self.parse_time parameter_name, time_string
    if time_string.nil?
      raise DateError, "#{parameter_name} is required"
    end

    unless time_string.match ISO8601_REGEXP
      raise DateError, "#{parameter_name} must be ISO 8601 datetime with timezone, optional punctuation, optional fractional seconds (1-3 digits)"
    end

    time = begin
             Time.parse( time_string ).utc
           rescue ArgumentError => e
             raise DateError, "#{parameter_name}: not a parsable time"
           end

    unless time >= EARLIEST_TIME and time <= LATEST_TIME
      raise DateError, "#{parameter_name} must be between #{EARLIEST_TIME.iso8601(3)} and #{LATEST_TIME.iso8601(3)} inclusive"
    end

    time
  end

  def self.log_environment
    LOG.info "SUBLEDGER_STORE: #{STORE}"
    LOG.info "SUBLEDGER_ENV:   #{ENV}"
  end
end

require 'subledger/store'
