module Subledger
  module Domain
    class AccountError < Error; end

    class Account

      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable
      include Roles::Updatable

      include Roles::Collectable

      include Roles::Activatable
      include Roles::Archivable

      include Roles::Restable

      attr_reader   :book
      attr_accessor :normal_balance

      def self.post_keys
        [ :description, :reference, :normal_balance ]
      end

      def self.patch_keys
        [ :id, :description, :reference, :normal_balance, :version ]
      end

      def self.root_klass
        Account
      end

      def self.sub_klasses
        [ active_klass, archived_klass ]
      end

      def self.active_klass
        ActiveAccount
      end

      def self.archived_klass
        ArchivedAccount
      end

      def self.first_and_last_line args
        store  = args[:store]
        client = args[:client]

        id = args[:id]

        if id.nil?
          raise AccountError, 'cannot get first and last lines without an :id'
        end

        args.merge! :account => client.accounts( :id => id )

        store.first_and_last_line args
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @book           = args[:book]
        @normal_balance = args[:normal_balance]
      end

      def line args
        client.posted_lines( args.merge! :account => self )
      end

      def lines args={ }, &block
        args.merge! :action       => args[:action]       || :ending,
                    :state        => args[:state]        || :posted,
                    :effective_at => args[:effective_at] || Time.now.utc,
                    :account      => self

        client.account_lines.collect args, &block
      end

      def balance args
        at = args[:at]

        unless at.kind_of? Time
          raise AccountError, ':at must be a Time'
        end

        store.account_balance :store   => store,
                              :client  => client,
                              :account => self,
                              :at      => at.utc
      end

      def first_and_last_line
        client.accounts.first_and_last_line :id => self.id
      end

      class Entity < Grape::Entity
        root 'accounts', 'account'

        expose :id,             :documentation => { :type => 'string',  :desc => 'account ID' }

        expose :book,           :format_with   => :id,
                                :documentation => { :type => 'string',  :desc => 'book ID' }

        expose :description,    :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,      :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :normal_balance, :format_with   => :normal_balance,
                                :documentation => { :type => 'string',  :desc => 'normal balance type' }

        expose :version,        :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      ACCEPTABLE_NORMAL_BALANCES = [ Debit, Credit ]

      def self.raise_unless_creatable args

        book = args[:book]

        if book.nil? or not book.kind_of? Book
          raise AccountError, ':book is required and must be a Book'
        elsif UUID.invalid? book.id
          raise AccountError, ':book must have a valid :id'
        end

        normal_balance = args[:normal_balance]

        if normal_balance.nil? or not ACCEPTABLE_NORMAL_BALANCES.include? normal_balance
          raise AccountError, ":normal_balance is required and must be one of #{ACCEPTABLE_NORMAL_BALANCES.join(', ')}"
        end
      end
    end

    class ArchivedAccount < Account
      class Entity < Account::Entity
        root 'archived_accounts', 'archived_account'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveAccount < Account
      class Entity < Account::Entity
        root 'active_accounts', 'active_account'
      end

      def self.sub_klasses
        [ active_klass ]
      end
    end
  end
end
