module Subledger
  module Domain
    class BalanceError < Error; end

    class Balance

      # Implements Valuable directly: #value, #amount, #+ and #==
      # Implements Attributable directly: #attributes

      include Domain

      include Roles::Restable

      def initialize args={ :debit_value => Zero.new, :credit_value => Zero.new }
        if args[:debit_value].nil? or args[:credit_value].nil?
          raise BalanceError, ':debit_value and :credit_value required'
        end

        @debit_amount  = args[:debit_value].amount
        @credit_amount = args[:credit_value].amount
      end

      def debit_value
        debit_amount.zero? ? Zero.new : Debit.new( debit_amount )
      end

      def credit_value
        credit_amount.zero? ? Zero.new : Credit.new( credit_amount )
      end

      def balanced?
        value == Zero.new
      end

      def value
        if debit_amount == credit_amount
          Zero.new
        elsif debit_amount > credit_amount
          Debit.new( debit_amount - credit_amount )
        else
          Credit.new( credit_amount - debit_amount )
        end
      end

      def amount
        value.amount
      end

      def + other
        klass = other.class

        debit_amount  = @debit_amount
        credit_amount = @credit_amount

        if klass == Balance
          debit_amount  += other.debit_value.amount
          credit_amount += other.credit_value.amount
        elsif klass == Debit
          debit_amount += other.amount
        elsif klass == Credit
          credit_amount += other.amount
        elsif klass == Zero
          return self.dup
        elsif other.kind_of? Line
          return self + other.value
        else
          raise BalanceError, "cannot add #{other.class.name} to a Balance"
        end

        self.class.new :debit_value  => debit_amount.zero?  ? Zero.new : Debit.new(debit_amount),
                       :credit_value => credit_amount.zero? ? Zero.new : Credit.new(credit_amount)
      end

      def == other
        debit_value == other.debit_value and credit_value == other.credit_value
      end

      def attributes
        { :debit_value  => debit_value,
          :credit_value => credit_value,
          :value        => value }
      end

      class Entity < Grape::Entity
        root 'balances', 'balance'

        expose :debit_value,  :format_with   => :value,
                              :documentation => { :type => 'string',  :desc => 'Total debits' }

        expose :credit_value, :format_with   => :value,
                              :documentation => { :type => 'string',  :desc => 'Total credits' }

        expose :value,        :format_with   => :value,
                              :documentation => { :type => 'string',  :desc => 'Total combined debits and credits' }
      end

      private

      attr_reader :debit_amount, :credit_amount
    end
  end
end
