module Subledger
  module Domain
    class CategoryError < Error; end

    class Category

      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable
      include Roles::Updatable

      include Roles::Collectable

      include Roles::Activatable
      include Roles::Archivable

      include Roles::Restable

      attr_reader   :book
      attr_accessor :normal_balance

      def self.post_keys
        [ :description, :reference, :normal_balance ]
      end

      def self.patch_keys
        [ :id, :description, :reference, :normal_balance, :version ]
      end

      def self.root_klass
        Category
      end

      def self.sub_klasses
        [ active_klass, archived_klass ]
      end

      def self.active_klass
        ActiveCategory
      end

      def self.archived_klass
        ArchivedCategory
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @book           = args[:book]
        @normal_balance = args[:normal_balance]
      end

      def accounts &block
        begin
          store.collect_accounts_for_category self, &block
        rescue Store::CollectError => e
          raise CategoryError, e
        end
      end

      class Entity < Grape::Entity
        root 'categories', 'category'

        expose :id,             :documentation => { :type => 'string',  :desc => 'category ID' }

        expose :book,           :format_with   => :id,
                                :documentation => { :type => 'string',  :desc => 'book ID' }

        expose :description,    :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,      :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :normal_balance, :format_with   => :normal_balance,
                                :documentation => { :type => 'string',  :desc => 'normal balance type' }

        expose :version,        :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      ACCEPTABLE_NORMAL_BALANCES = [ Debit, Credit ]

      def self.raise_unless_creatable args

        book = args[:book]

        if book.nil? or not book.kind_of? Book
          raise CategoryError, ':book is required and must be a Book'
        elsif UUID.invalid? book.id
          raise CategoryError, ':book must have a valid :id'
        end

        normal_balance = args[:normal_balance]

        if normal_balance.nil? or not ACCEPTABLE_NORMAL_BALANCES.include? normal_balance
          raise CategoryError, ":normal_balance is required and must be one of #{ACCEPTABLE_NORMAL_BALANCES.join(', ')}"
        end
      end
    end

    class ArchivedCategory < Category
      class Entity < Category::Entity
        root 'archived_categories', 'archived_category'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveCategory < Category
      class Entity < Category::Entity
        root 'active_categories', 'active_category'
      end

      def self.sub_klasses
        [ active_klass ]
      end

      def attach args
        account = args[:account]

        begin
          store.attach_account_to_category :category => self,
                                           :account  => account
        rescue Store::AttachError => e
          raise CategoryError, e
        end

        account
      end

      def detach args
        account = args[:account]

        begin
          store.detach_account_from_category :category => self,
                                             :account  => account
        rescue Store::DetachError => e
          raise CategoryError, e
        end

        account
      end
    end
  end
end
