module Subledger
  module Domain
    class ControlError < Error; end

    class Control

      include Domain

      include Roles::Attributable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Creatable
      include Roles::Readable
      include Roles::Collectable
      include Roles::Activatable
      include Roles::Archivable
      include Roles::Restable

      attr_reader :identity, :verbs, :path, :mode

      def self.root_klass
        Control
      end

      def self.post_keys
        [ :id, :identity, :verbs, :path, :mode ]
      end

      def self.patch_keys
        [  ]
      end

      def self.sub_klasses
        [ active_klass, archived_klass ]
      end

      def self.active_klass
        ActiveControl
      end

      def self.archived_klass
        ArchivedControl
      end

      def initialize args
        identifiable args
        storable args

        @identity = args[:identity]
        @verbs    = args[:verbs]
        @path     = args[:path]
        @mode     = args[:mode]
      end

      # def revoke
      #   @revoked = true
      #
      #   self.update
      #
      #   self
      # end

      # def revoked?
      #   @revoked == true
      # end

      # def is verb, url
      #   verb == @verbs and url == @path
      # end

      def match? verb, url
        @verbs.split('|').include?(verb) and match_wildcard(url, @path)
      end

      class Entity < Grape::Entity
        root 'controls', 'control'

        expose :id,       :documentation => { :type => 'string', :desc => 'control ID' }

        expose :identity, :format_with   => :id,
                          :documentation => { :type => 'string', :desc => 'identity ID' }

        expose :verbs,    :documentation => { :type => 'string', :desc => 'HTTP verbs' }

        expose :path,     :documentation => { :type => 'string', :desc => 'HTTP path' }

        expose :mode,     :documentation => { :type => 'string', :desc => 'allow or deny' }
      end

      private

      def match_wildcard actual, authorized
        actual     = strip_version actual
        authorized = strip_version authorized

        regx = authorized.gsub(/\*(.)/, '[^\1]*\1').gsub(/\*$/, '.*')

        actual == authorized or ( actual =~ /^#{regx}$/ ) == 0
      end

      def strip_version path
        path[3..-1]
      end

      ACCEPTABLE_VERBS = [ 'POST', 'GET', 'PATCH' ]
      ACCEPTABLE_MODES = [ :allow, :deny ]

      def self.raise_unless_creatable args

        identity = args[:identity]

        if identity.nil? or not identity.kind_of? Identity
          raise ControlError, ':identity is required and must be an Identity'
        elsif UUID.invalid? identity.id
          raise ControlError, ':identity must have a valid :id'
        end

        # TODO :verb(s) should be a domain object

        verbs = args[:verbs]

        if verbs.nil?
          raise ControlError, ':verbs is required'
        end

        verbs.split('|').each do |verb|
          unless ACCEPTABLE_VERBS.include? verb
            raise ControlError, ":verbs must contain one or more of #{ACCEPTABLE_VERBS.join(', ')} separated by |"
          end
        end

        # TODO :path should be a domain object

        path = args[:path]

        if path.nil?
          raise ControlError, ':path is required'
        end

        api_version = path[1,2]

        unless api_version == API_VERSION
          raise ControlError, ":path must begin with /#{API_VERSION}/"
        end

        path[4,].split( '/' ).each do |part|
          unless part =~ /[*_0-9A-Za-z]+/
            raise ControlError, 'invalid :path'
          end
        end

        mode = args[:mode]

        if mode.nil? or not ACCEPTABLE_MODES.include? mode
          raise ControlError, ":mode must contain one or more of :allow, :deny}"
        end
      end
    end

    class ArchivedControl < Control
      class Entity < Control::Entity
        root 'archived_controls', 'archived_control'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveControl < Control
      class Entity < Control::Entity
        root 'active_controls', 'active_control'
      end

      def self.sub_klasses
        [ active_klass ]
      end
    end
  end
end
