require 'mail' # this can be eliminated when the gem below accepts our PR
require 'valid_email/validate_email'

module Subledger
  module Domain
    class IdentityError < Error; end

    class Identity

      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable
      include Roles::Updatable

      include Roles::Activatable
      include Roles::Archivable

      include Roles::Restable

      attr_accessor :email

      def self.post_keys
        [ :id, :email, :description, :reference ]
      end

      def self.patch_keys
        [ :id, :email, :description, :reference, :version ]
      end

      def self.root_klass
        Identity
      end

      def self.sub_klasses
        [ archived_klass, active_klass ]
      end

      def self.archived_klass
        ArchivedIdentity
      end

      def self.active_klass
        ActiveIdentity
      end

      def self.create args
        validate_creatability args

        identity = args[:client].send active_klass.collection_name, args

        args[:store].create_identity identity
      end

      def create
        identity, key = self.class.create self.attributes

        initialize identity.attributes

        return self, key
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @email = args[:email]
      end

      # def by_email email
      #   # TODO
      # end

      # def by_org org
      #   # TODO look in OrgIdentities table
      # end

      def authorize verb, url
        raise KeyError if @retired

        controls.each do |control|
          return true if control.match(verb, url) unless control.revoked?
        end

        false
      end

      # def grant args
      #   args[:client].controls.create args
      # end

      # def revoke args
      #   org_id = args[:org_id]
      #   verb   = args[:verb]
      #   url    = args[:url]
      #
      #   revoked = false
      #
      #   controls.each do |control|
      #     if control.is org_id, verb, url
      #       control.revoke
      #       revoked = true
      #     end
      #   end
      #
      #   raise IdentityError unless revoked
      #
      #   self
      # end

      class Entity < Grape::Entity
        root 'identities', 'identity'

        expose :id,          :documentation => { :type => 'string',  :desc => 'identity ID' }

        expose :email,       :documentation => { :type => 'string',  :desc => 'identity email' }

        expose :description, :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,   :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :version,     :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      def self.raise_unless_creatable args
        unless ValidateEmail.mx_valid? args[:email]
          raise IdentityError, ':email is required and must have an MX record for the domain'
        end
      end
    end

    class ArchivedIdentity < Identity
      class Entity < Identity::Entity
        root 'archived_identities', 'archived_identity'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveIdentity < Identity
      class Entity < Identity::Entity
        root 'active_identities', 'active_identity'
      end

      def self.sub_klasses
        [ active_klass ]
      end
    end
  end
end
