module Subledger
  module Domain
    class JournalEntryError < Error; end

    class JournalEntry

      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Timeable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable
      include Roles::Updatable

      include Roles::Progressable
      include Roles::Collectable

      include Roles::Restable

      attr_reader :org, :book, :post_delay

      attr_accessor :effective_at

      def self.post_keys
        [ :effective_at, :description, :reference ]
      end

      def self.patch_keys
        [ :id, :effective_at, :description, :reference, :version ]
      end

      def self.root_klass
        JournalEntry
      end

      def self.sub_klasses
        [ active_klass, archived_klass, posting_klass, posted_klass ]
      end

      def self.active_klass
        ActiveJournalEntry
      end

      def self.archived_klass
        ArchivedJournalEntry
      end

      def self.posting_klass
        PostingJournalEntry
      end

      def self.posted_klass
        PostedJournalEntry
      end

      def self.create_and_post args
        begin
          client = args[:client]

          active_journal_entry = client.active_journal_entry args

          JournalEntry.send :validate_creatability, active_journal_entry.attributes

          order = 1

          arg_lines = args[:lines]

          arg_lines.each do |line_args|
            if line_args[:description].nil?
              line_args.merge! :description => active_journal_entry.description
            end

            if line_args[:reference].nil?
              line_args.merge! :reference => active_journal_entry.reference
            end

            line_args.merge! :order => '%07.2f' % order

            order += 1
          end

          store = args[:store]

          active_lines = []

          arg_lines.each do |line_args|
            line_args.merge! :journal_entry => active_journal_entry

            line = client.active_line line_args.merge( :id => UUID.as_string )

            Line.send :raise_unless_create_and_postable, line.attributes
            Line.send :validate_creatability_modules, line.attributes

            active_lines << line
          end

          active_journal_entry.send :set_active_lines, active_lines

          store.create_and_post(
                  :active_journal_entry  => active_journal_entry,
                  :active_lines          => active_lines,
                  :posting_journal_entry => client.posting_journal_entry( { } ) )

        rescue Exception => e
          raise JournalEntryError, "Cannot create and post: #{e}"
        end
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @org          = args[:org]
        @book         = args[:book]
        @effective_at = utc_or_nil args[:effective_at]
        @post_delay   = args[:post_delay] || 0

        @reason = args[:reason] if args.has_key? :reason
      end

      def line args
        client.lines args.merge( :journal_entry => self )
      end

      def lines args={ }, &block
        return @active_lines if @active_lines

        args.merge! :action        => args[:action] || :starting,
                    :state         => args[:state]  || line_state,
                    :order         => args[:order]  || '0',
                    :journal_entry => self

        client.lines.collect args, &block
      end

      def balance
        if @active_lines
          active_lines_balance = client.balance
          @active_lines.each { |line| active_lines_balance += line }
          active_lines_balance
        else
          store.journal_entry_balance :store         => store,
                                      :client        => client,
                                      :journal_entry => self,
                                      :state         => line_state
        end
      end

      def balanced?
        balance.balanced?
      end

      class Entity < Grape::Entity
        root 'journal_entries', 'journal_entry'

        expose :id,           :documentation => { :type => 'string',  :desc => 'journal entry ID' }

        expose :book,         :format_with   => :id,
                              :documentation => { :type => 'string',  :desc => 'book ID' }

        expose :effective_at, :format_with   => :time,
                              :documentation => { :type => 'string',  :desc => 'effective at date' }

        expose :description,  :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,    :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :version,      :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      def self.raise_unless_creatable args
        book = args[:book]

        if book.nil? or not book.kind_of? Book
          raise JournalEntryError, ':book is required and must be a Book'
        elsif UUID.invalid? book.id
          raise JournalEntryError, ':book must have a valid :id'
        end

        effective_at = args[:effective_at]

        if effective_at.nil? or not effective_at.kind_of? Time
          raise JournalEntryError, ':effective_at is required and must be a Time'
        end
      end

      def line_state
        :active
      end
    end

    class ArchivedJournalEntry < JournalEntry
      include Roles::Activatable
      include Roles::Archivable

      class Entity < JournalEntry::Entity
        root 'archived_journal_entries', 'archived_journal_entry'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveJournalEntry < JournalEntry

      include Roles::Postable
      include Roles::Activatable
      include Roles::Archivable

      attr_reader :reason

      class Entity < JournalEntry::Entity
        root 'active_journal_entries', 'active_journal_entry'

        expose :reason
      end

      def self.sub_klasses
        [ active_klass ]
      end

      def create_line args
        args.merge! :client        => client,
                    :journal_entry => self,
                    :effective_at  => effective_at

        Line.raise_on_duplicate_orders args

        client.active_lines.validate_creatability args

        active_line = client.active_lines args

        store.create_line :line => active_line
      end

      private

      def set_active_lines active_lines
        @active_lines = active_lines
      end

      def set_reason reason
        @reason = reason
      end
    end

    class PostingJournalEntry < JournalEntry
      include Roles::Postable

      class Entity < JournalEntry::Entity
        root 'posting_journal_entries', 'posting_journal_entry'
      end

      def self.sub_klasses
        [ posting_klass ]
      end

      def self.update args
        raise JournalEntryError, 'posting journal entries are not updatable'
      end

      def update
        self.class.update( { } )
      end

      def balance
        raise JournalEntryError, 'cannot call balance on a PostingJournalEntry'
      end
    end

    class PostedJournalEntry < JournalEntry
      include Roles::Postable

      class Entity < JournalEntry::Entity
        root 'posted_journal_entries', 'posted_journal_entry'
      end

      def self.sub_klasses
        [ posted_klass ]
      end

      def self.update args
        raise JournalEntryError, 'posted journal entries are not updatable'
      end

      def update
        self.class.update( { } )
      end

      private

      def line_state
        :posted
      end
    end
  end
end
