module Subledger
  module Domain
    class LineError < Error; end

    class Line

      include Domain::Value
      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Timeable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable

      include Roles::Collectable

      include Roles::Restable

      attr_reader   :journal_entry
      attr_accessor :account, :value, :order

      def self.root_klass
        Line
      end

      def self.sub_klasses
        [ archived_klass, active_klass, posted_klass ]
      end

      def self.archived_klass
        ArchivedLine
      end

      def self.active_klass
        ActiveLine
      end

      def self.posted_klass
        PostedLine
      end

      def self.update args
        raise_on_duplicate_orders args

        super
      end

      def self.specialized_raise_unless_creatable args
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @journal_entry = args[:journal_entry]
        @account       = args[:account]
        @value         = args[:value]
        @order         = args[:order]

        specialized_initialization args
      end

      def amount
        value.amount
      end

      class Entity < Grape::Entity
        root 'lines', 'line'

        expose :id,            :documentation => { :type => 'string',  :desc => 'line ID' }

        expose :journal_entry, :format_with   => :id,
                               :documentation => { :type => 'string',  :desc => 'journal entry ID' }

        expose :account,       :format_with   => :id,
                               :documentation => { :type => 'string',  :desc => 'account ID' }

        expose :description,   :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,     :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :value,         :format_with   => :value,
                               :documentation => { :type => 'string',  :desc => 'value' }

        expose :order,         :documentation => { :type => 'string',  :desc => 'order' }

        expose :version,       :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      def self.specialized_keys
        []
      end

      def self.raise_on_duplicate_orders args
        client = args[:client]

        journal_entry = client.journal_entry :id => args[:journal_entry].id

        order = BigDecimal args[:order]

        journal_entry.lines.each do |line|
          if order == BigDecimal( line.order )
            raise LineError, "lines must have unique orders: #{args[:orders]}"
          end
        end
      end

      def self.raise_unless_creatable args
        raise_on_duplicate_orders args

        raise_unless_create_and_postable args

        journal_entry = args[:journal_entry]

        if UUID.invalid? journal_entry.id
          raise LineError, ':journal_entry must have a valid :id'
        end
      end

      def self.raise_unless_create_and_postable args
        journal_entry = args[:journal_entry]

        if journal_entry.nil? or not journal_entry.kind_of? JournalEntry
          raise LineError, ':journal_entry is required and must be a JournalEntry'
        end

        account = args[:account]

        if account.nil? or not account.kind_of? Account
          raise LineError, ':account is required and must be an Account'
        elsif UUID.invalid? account.id
          raise LineError, ':account must have a valid :id'
        end

        value = args[:value]

        if value.nil? or not value.kind_of? Value
          raise LineError, ':value is required and must be a Value'
        end

        order = args[:order]

        if order.nil? or
           order !~ /^\d{1,4}[.]?\d{0,2}$/
          raise LineError, ':order is required and in the form [###]#[.##]'
        end

        bd_order = BigDecimal order

        if bd_order <= 0 or bd_order >= 10_000
          raise LineError, ':order must be > 0 and < 10,000'
        end

        specialized_raise_unless_creatable args
      end

      def specialized_initialization args
      end
    end

    class ArchivedLine < Line

      include Roles::Updatable
      include Roles::Archivable
      include Roles::Activatable

      class Entity < Line::Entity
        root 'archived_lines', 'archived_line'
      end

      def self.sub_klasses
        [ archived_klass ]
      end

      def self.post_keys
        [ :account, :journal_entry, :description, :reference, :value, :order ]
      end

      def self.patch_keys
        [ :id, :account, :description, :reference, :value, :order, :version ]
      end

      def posted?
        false
      end
    end

    class ActiveLine < Line

      include Roles::Updatable
      include Roles::Archivable
      include Roles::Activatable

      class Entity < Line::Entity
        root 'active_lines', 'active_line'
      end

      def self.sub_klasses
        [ active_klass ]
      end

      def self.post_keys
        [ :account, :journal_entry, :description, :reference, :value, :order ]
      end

      def self.patch_keys
        [ :id, :account, :description, :reference, :value, :order, :version ]
      end

      def posted?
        false
      end
    end

    class PostedLineError < Error; end

    class PostedLine < Line
      attr_reader :effective_at, :posted_at, :balance

      class Entity < Line::Entity
        root 'posted_lines', 'posted_line'

        expose :effective_at, :format_with   => :time,
                              :documentation => { :type => 'string',  :desc => 'effective time' }

        expose :posted_at,    :format_with   => :time,
                              :documentation => { :type => 'string',  :desc => 'posted time' }

        expose :balance,      :format_with   => :balance
      end

      def self.sub_klasses
        [ posted_klass ]
      end

      def posted?
        true
      end

      private

      attr_writer :balance

      def self.specialized_raise_unless_creatable args
        effective_at = args[:effective_at]

        if effective_at.nil? or not effective_at.kind_of? Time
          raise PostedLineError, ':effective_at is required and must be a Time'
        end

        if args[:posted_at].nil?
          raise PostedLineError, ':posted_at is required'
        end
      end

      def specialized_initialization args
        @effective_at = utc_or_nil args[:effective_at]
        @posted_at    = utc_or_nil args[:posted_at]
        @balance      = args[:balance]
      end
    end
  end
end
