module Subledger
  module Domain
    class ReportError < Error; end

    class Report

      include Domain

      include Roles::Attributable
      include Roles::Describable
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Versionable

      include Roles::Creatable
      include Roles::Readable
      include Roles::Updatable

      include Roles::Collectable

      include Roles::Activatable
      include Roles::Archivable

      include Roles::Restable

      attr_reader :book

      def self.post_keys
        [ :description, :reference ]
      end

      def self.patch_keys
        [ :id, :description, :reference, :version ]
      end

      def self.root_klass
        Report
      end

      def self.sub_klasses
        [ active_klass, archived_klass ]
      end

      def self.active_klass
        ActiveReport
      end

      def self.archived_klass
        ArchivedReport
      end

      def initialize args
        describable args
        identifiable args
        storable args
        versionable args

        @book = args[:book]
      end

      class Entity < Grape::Entity
        root 'reports', 'report'

        expose :id,          :documentation => { :type => 'string',  :desc => 'report ID' }

        expose :book,        :format_with   => :id,
                             :documentation => { :type => 'string',  :desc => 'book ID' }

        expose :description, :documentation => { :type => 'string',  :desc => 'description' }

        expose :reference,   :documentation => { :type => 'string',  :desc => 'reference URI' }

        expose :version,     :documentation => { :type => 'integer', :desc => 'version' }
      end

      private

      def self.raise_unless_creatable args

        book = args[:book]

        if book.nil? or not book.kind_of? Book
          raise BookError, ':book is required and must be a Book'
        elsif UUID.invalid? book.id
          raise BookError, ':book must have a valid :id'
        end
      end

      def structure
        store.report_structure_for self
      end
    end

    class ArchivedReport < Report
      class Entity < Report::Entity
        root 'archived_reports', 'archived_report'
      end

      def self.sub_klasses
        [ archived_klass ]
      end
    end

    class ActiveReport < Report
      class Entity < Report::Entity
        root 'active_reports', 'active_report'
      end

      def self.sub_klasses
        [ active_klass ]
      end

      def categories &block
        begin
          store.collect_categories_for_report self, &block
        rescue Store::CollectError => e
          raise ReportError, e
        end
      end

      def attach args
        validate_attach args

        category = args[:category]
        parent   = args[:parent]

        if category.nil? or category.id.nil?
          raise ReportError, ':category is required and must have an id'
        end

        if not parent.nil? and parent.id.nil?
          raise ReportError, ':parent must have an id'
        end

        begin
          store.attach_category_to_report :report   => self,
                                          :category => category,
                                          :parent   => parent
        rescue Store::AttachError => e
          raise ReportError, e
        end

        category
      end

      def detach args
        category = args[:category]

        if category.nil? or category.id.nil?
          raise ReportError, ':category is required and must have an id'
        end

        begin
          store.detach_category_from_report :report   => self,
                                            :category => category
        rescue Store::DetachError => e
          raise ReportError, e
        end

        category
      end

      def render args
        at = args[:at]

        if at.nil? or not at.kind_of? Time
          raise ReportError, ':at is required and must be a Time'
        end

        building_report_rendering = client.building_report_rendering :effective_at => at,
                                                                     :report       => self

        store.render args.merge! :building_report_rendering => building_report_rendering
      end

      private

      def add_categories_to report_hash, sub_structure, effective_at
        categories_hash = report_hash[:categories] = { }

        categories_balance = Balance.new

        sub_structure.each do |category, category_hash|
          categories_hash[category.id] = category.serializable_hash

          category_balance = add_categories_to categories_hash[category.id], category_hash, effective_at

          category_balance += add_accounts_to categories_hash[category.id], category, effective_at

          categories_hash[category.id][:balance] = category_balance.serializable_hash

          categories_balance += category_balance
        end

        categories_balance
      end

      def add_accounts_to report_category_hash, category, effective_at
        accounts_hash = report_category_hash[:accounts] = { }

        accounts_balance = Balance.new

        category.accounts.each do |account|
          accounts_hash[account.id] = account.serializable_hash

          account_balance = add_balance_to accounts_hash[account.id], account, effective_at

          accounts_balance += account_balance
        end

        accounts_balance
      end

      def add_balance_to report_category_account_hash, account, effective_at
        balance = account.balance :at => effective_at

        report_category_account_hash[:balance] = balance.serializable_hash

        balance
      end

      def validate_attach args
        if UUID.invalid? id
          raise ReportError, ':report must have a valid :id'
        end

        category = args[:category]

        if category.nil? or not category.kind_of? Category
          raise ReportError, ':category is required and must be a Category'
        elsif UUID.invalid? category.id
          raise ReportError, ':category must have a valid :id'
        end

        parent = args[:parent]

        unless parent.nil?
          if not parent.kind_of? Category
            raise ReportError, ':parent must be a Category'
          elsif UUID.invalid? parent.id
            raise ReportError, ':parent must have a valid :id'
          end
        end
      end
    end
  end
end
