module Subledger
  module Domain
    class ReportRenderingError < Error; end

    class ReportRendering

      include Domain

      include Roles::Attributable
      include Roles::DescribableReportRendering
      include Roles::Identifiable
      include Roles::Storable
      include Roles::Timeable

      include Roles::Creatable
      include Roles::Readable

      include Roles::Collectable

      include Roles::Restable

      attr_reader :book, :report, :effective_at, :rendered_at,
                  :balanced_accounts, :total_accounts

      def self.post_keys
        [ :description, :reference, :report, :effective_at ]
      end

      def self.root_klass
        ReportRendering
      end

      def self.sub_klasses
        [ building_klass, completed_klass ]
      end

      def self.building_klass
        BuildingReportRendering
      end

      # TODO create uses active_klass, should use creation_klass
      def self.active_klass
        building_klass
      end

      def self.completed_klass
        CompletedReportRendering
      end

      def initialize args
        describable_report_rendering args
        identifiable args
        storable args

        @book              = args[:book]
        @report            = args[:report]
        @effective_at      = round args[:effective_at]
        @rendered_at       = utc_or_now args[:rendered_at]

        @balanced_accounts = args[:balanced_accounts] || 0
        @total_accounts    = args[:total_accounts]    || 0

        specialized_initialization args
      end

      def progress
        if total_accounts.zero?
          0
        else
          ( ( balanced_accounts.to_f / total_accounts ) * 99 ).to_i
        end
      end

      class Entity < Grape::Entity
        root 'report_renderings', 'report_rendering'

        expose :id
        expose :book,           :format_with => :id
        expose :report,         :format_with => :id
        expose :description
        expose :reference
        expose :effective_at,   :format_with => :time_full_second
        expose :total_accounts, :format_with => :integer
        expose :progress,       :format_with => :progress
        expose :rendered_at,    :format_with => :time
      end

      private

      def self.raise_unless_creatable args

        book = args[:book]

        if book.nil? or not book.kind_of? Book
          raise ReportRenderingError, ':book is required and must be a Book'
        elsif UUID.invalid? book.id
          raise ReportRenderingError, ':book must have a valid :id'
        end

        report = args[:report]

        if report.nil? or not report.kind_of? Report
          raise ReportRenderingError, ':report is required and must be a Report'
        elsif UUID.invalid? report.id
          raise ReportRenderingError, ':report must have a valid :id'
        end

        effective_at = args[:effective_at]

        if effective_at.nil? or not effective_at.kind_of? Time
          raise ReportRenderingError, ':effective_at is required and must be a Time'
        end
      end

      def round effective_at
        Time.at( utc_or_nil( effective_at ).to_i ).utc
      end

      TOTAL_ACCOUNTS_SEMAPHORE = Mutex.new

      def increase_total_accounts_by count
        TOTAL_ACCOUNTS_SEMAPHORE.synchronize do
          @total_accounts += count
        end
      end

      BALANCED_ACCOUNTS_SEMAPHORE = Mutex.new

      def increase_balanced_accounts_by count
        BALANCED_ACCOUNTS_SEMAPHORE.synchronize do
          @balanced_accounts += count
        end
      end

      def specialized_initialization args
      end
    end

    class BuildingReportRendering < ReportRendering
      class Entity < ReportRendering::Entity
        root 'building_report_renderings', 'building_report_rendering'
      end

      def self.sub_klasses
        [ building_klass ]
      end

      # TODO this should probably move to Domain module

      def to_json
        MultiJson.dump serializable_hash
      end

      private

      def complete args
        store.send :become, :becomable  => self,
                            :klass      => self.class.completed_klass,
                            :extra_args => args
      end
    end

    class CompletedReportRendering < ReportRendering
      attr_reader :completed_at, :categories, :balance, :warnings

      def self.sub_klasses
        [ completed_klass ]
      end

      def progress
        100
      end

      def to_json
        @json ||= if location[0..4] == 's3://'
                    ExceptionHandler.new( :name => 'completed_report_rendering to_json' ).with_retry do
                      s3_bucket.objects[id].read
                    end
                  elsif location[0..6] == 'file://'
                    File.open( location[7..-1], 'r' ).read
                  else
                    raise ReportRenderingError, "unknown URL scheme: #{location}"
                  end
      end

      def categories
        @categories ||= as_hash['categories']
      end

      def balance
        balance_hash = { 'balance' => as_hash['balance'] }

        @balance ||= Rest.to_balance balance_hash, client
      end

      def warnings
        @warnings ||= as_hash['warnings']
      end

      class Entity < ReportRendering::Entity
        root 'completed_report_renderings', 'completed_report_rendering'

        expose :completed_at, :format_with => :time
        expose :categories
        expose :balance,      :format_with => :balance
        expose :warnings
      end

      private

      attr_reader :location

      def s3_bucket
        @s3_bucket ||= Store::Aws.s3_client.buckets[RENDERED_REPORTS_BUCKET]
      end

      def as_hash
        @hash ||= MultiJson.load to_json
      end

      def specialized_initialization args
        @completed_at = utc_or_nil args[:completed_at]
        @location     = args[:location]
        @json         = args[:json]

        @balance      = nil
        @categories   = nil
        @hash         = nil
        @warnings     = nil
      end
    end
  end
end
