module Subledger
  module Domain
    module Roles
      module Collectable
        def self.included base
          base.extend CollectableClass
        end

        module CollectableClass

          def collect args, &block
            limit = args[:limit] ||= 25 unless block_given?

            validate args

            store           = args[:store]
            client          = args[:client]
            anchor          = args[:anchor]
            action          = args[:action]
            collection_name =  CollectionName.with_state( args.
                                                            merge :klass => self )

            args[:collection_name] = collection_name

            unless anchor
              anchor = args[:anchor] = client.send collection_name, args

              anchor.read if [:preceding, :following].include? action
            end

            validate_criterion args

            if block_given?
              collect_with_block args, &block
            else
              store.collect args
            end
          end

          private

          ACTION_MAP = { :before    => :preceding,
                         :ending    => :preceding,
                         :starting  => :following,
                         :after     => :following,
                         :preceding => :preceding,
                         :following => :following }

          MAX_LIMIT = 100

          def collect_with_block args
            store  = args[:store]
            anchor = args[:anchor]
            action = args[:action]
            limit  = args[:limit]

            next_action = ACTION_MAP[ action ]

            total = 0

            store_limit = first_limit limit

            while not ( finished_collecting? limit, total ) and
                  not ( collected = store.collect(
                                      args.merge!(
                                        :anchor => anchor,
                                        :action => action,
                                        :limit  => store_limit ) ) ).empty?

              collected.each do |item|
                yield item

                total += 1
              end

              action      = next_action
              anchor      = collected.last
              store_limit = next_limit limit, total
            end
          end

          def finished_collecting? limit, total
            not limit.nil? and total >= limit
          end

          def first_limit limit
            ( limit.nil? or limit > MAX_LIMIT ) ? MAX_LIMIT : limit
          end

          def next_limit limit, total
            limit.nil? ? MAX_LIMIT : calculated_limit( limit, total )
          end

          def calculated_limit limit,total
            new_limit = limit - total

            new_limit < MAX_LIMIT ? new_limit : MAX_LIMIT
          end

          def validate args
            validate_action    args
            validate_limit     args
            validate_state     args
          end

          ACTIONS = [ :before,
                      :ending,
                      :starting,
                      :after,
                      :preceding,
                      :following ]

          def validate_action args
            unless ACTIONS.include? args[:action]
              raise CollectableError, ':action must be :before, :ending, :starting, :after, :preceding, or :following'
            end
          end

          def validate_limit args
            limit = args[:limit]

            if not limit.nil? and limit < 1
              raise CollectableError, ':limit must be greater than zero or nil'
            end
          end

          def validate_state args
            state = args[:state]

            if state.nil?
              raise CollectableError, ':state is required'
            end
          end

          COLLECTABLE_BY  = { :building_report_renderings  => :rendered_at,
                              :completed_report_renderings => :rendered_at,

                              :archived_accounts           => :description,
                              :active_accounts             => :description,
                              :archived_books              => :description,
                              :active_books                => :description,
                              :archived_categories         => :description,
                              :active_categories           => :description,
                              :archived_reports            => :description,
                              :active_reports              => :description,

                              :account_lines               => :effective_at,
                              :active_journal_entries      => :effective_at,
                              :archived_journal_entries    => :effective_at,
                              :posting_journal_entries     => :effective_at,
                              :posted_journal_entries      => :effective_at,

                              :archived_lines              => :order,
                              :active_lines                => :order,
                              :posted_lines                => :order,

                              :archived_controls           => :id,
                              :active_controls             => :id,
                              :archived_keys               => :id,
                              :active_keys                 => :id }

          def validate_criterion args
            anchor          = args[ :anchor ]
            collection_name = args[ :collection_name ]
            required_field  = COLLECTABLE_BY[ collection_name ]

            if anchor.send( required_field ).nil?
              raise CollectableError, "#{required_field.inspect} is required"
            end
          end
        end
      end
    end
  end
end
